
# cmm_comp.py

from __future__ import print_function

from ckit_parser import *

import copy

# --------------------------------------------------------------------------

class Compiler (Parser) :
   def __init__ (self) :
       super (Compiler, self).__init__ ()

       self.global_scope = CmmNamespaceDecl ()
       self.global_scope.item_name = ""
       self.global_scope.qual_name = ""
       self.global_scope.item_dict = { }
       self.global_scope.item_list = [ ]
       self.display = [ self.global_scope ]

   # scope

   def enter (self, item, also_mark_defn = True) :
       top = self.display [-1]
       top.item_dict [item.item_name] = item
       top.item_list.append (item)
       item.item_context = top
       if top.qual_name == "" :
          item.qual_name = item.item_name
       else :
          item.qual_name = top.qual_name + "." + item.item_name
       if also_mark_defn :
          self.markDefn (item)

   def openScope (self, item, already_entered = False) :
       if not already_entered :
          self.enter (item)
       if not hasattr (item, "item_dict") :
          item.item_dict = { }
       if not hasattr (item, "item_list") :
          item.item_list = [ ]
       self.display.append (item)

   def closeScope (self) :
       self.display.pop ()

   def lookup (self, name) :
       result = None
       inx = len (self.display) - 1
       while result == None and inx >= 0 :
          result = self.display [inx].item_dict.get (name)
          inx = inx - 1
       return result

   def getName (self, name) :
       if name == None :
          return "" # !?
       else :
          return name.id

   # namespace

   def open_namespace (self, module) :
       module.item_name = self.getName (module.name)
       module.item_expand = True
       module.icon = "module"
       self.markOutline (module)
       self.openScope (module)
       self.openCompletion (module, following = True)

   def close_namespace (self, module) :
       self.closeCompletion (module)
       self.closeScope ()

   # class

   def open_class (self, cls) :
       cls.item_name = self.getName (cls.name)
       cls.item_expand = True
       cls.icon = "class"
       self.markOutline (cls) # navigator item
       # self.openRegion (cls) # region with background color
       self.openScope (cls) # scope for identifiers
       self.openCompletion (cls) # area with text completion
       # print ("class", cls.item_name)

   def close_class (self, cls) :
       self.closeCompletion (cls)
       self.closeScope ()
       # self.closeRegion ()

   # enum

   def open_enum (self, enum) :
       enum.item_name = self.getName (enum.name)
       enum.icon = "enum"
       self.markOutline (enum)
       # self.openRegion (enum)
       self.openScope (enum)
       self.openCompletion (enum)

   def on_enum_item (self, item) :
       item.item_name = self.getName (item.name)
       item.icon = "variable"
       self.enter (item)

   def close_enum (self, enum) :
       self.closeCompletion (enum)
       self.closeScope ()
       # self.closeRegion ()

   # variable

   def add_init_declarator (self, decl) :
       if decl.name != None :
          decl.item_name = self.getName (decl.name)
          decl.icon = "variable"
          self.enter (decl)

   # fuction

   #def open_function (self, simple_declaration) :
       #simple_item = simple_declaration.items [-1]
       #simple_item.icon = "function"
       #self.openRegion (simple_item)
       #self.openScope (simple_item, already_entered = True)
       #self.openCompletion (simple_item)
       #self.markOutline (simple_item)

   #def close_function (self, simple_declaration) :
       #simple_item = simple_declaration.items [-1]
       #self.closeCompletion (simple_item)
       #self.closeScope ()
       #self.closeRegion ()
       #simple_declaration.func_body.item_label = simple_item.item_name

   # predicates

   def is_parameter (self) :
       return True

   def is_template_arg (self) :
       return True

   def is_cont_template_arg (self) :
       return True

   def is_inner_abstract_declarator (self) :
       return True

   def is_inner_common_declarator (self) :
       return True

   # skip fuction body

   def parse_compound_stat (self) :
       result = CmmCompoundStat ()
       result.stat = result.compoundStat
       result.fileInx = self.tokenFileInx
       result.line = self.tokenLineNum
       result.column = self.tokenColNum
       result.pos = self.tokenByteOfs
       self.check ("{")
       level = 0
       while self.tokenText != "}" or level != 0 :
          if self.tokenText == "{" :
             level = level + 1
          if self.tokenText == "}" :
             level = level - 1
          self.nextToken ()
       self.check ("}")
       self.close_block (result)
       return result

   # skip parameter

   #def parse_parameter_declaration (self) :
       #level = 0
       #while (self.tokenText != "," and self.tokenText != ")") or level != 0 :
          #if self.tokenText == "(" :
             #level = level + 1
          #if self.tokenText == ")" :
             #level = level - 1
          #self.nextToken ()
       #return None

   # skip initializer

   def parameter_initialization (level) :
       self.check ("=")
       level = 0
       while (self.tokenText != "," and self.tokenText != ")") or level != 0 :
          if self.tokenText == "(" :
             level = level + 1
          if self.tokenText == ")" :
             level = level - 1
          self.nextToken ()
       return None

   def parse_initializer_item (self) :
       level = 0
       while (self.tokenText != "," and self.tokenText != ";") or level != 0 :
          if self.tokenText == "(" :
             level = level + 1
          if self.tokenText == ")" :
             level = level - 1
          self.nextToken ()
       return None

   def parse_enumerator_init (self) :
       level = 0
       while (self.tokenText != "," and self.tokenText != "}") or level != 0 :
          if self.tokenText == "(" :
             level = level + 1
          if self.tokenText == ")" :
             level = level - 1
          self.nextToken ()
       return None

   def parse_array_limit (self) :
       level = 0
       while self.tokenText != "]" and self.token != self.string_literal or level != 0 :
          if self.tokenText == "[" :
             level = level + 1
          if self.tokenText == "]" :
             level = level - 1
          self.nextToken ()
       return None

# --------------------------------------------------------------------------

# kate: indent-width 1; show-tabs true; replace-tabs true; remove-trailing-spaces all
