#!/usr/bin/env python

from __future__ import print_function

import os, sys, importlib, traceback

from util import use_qt5
if use_qt5 :
   from PyQt5.QtCore import *
   from PyQt5.QtGui import *
   from PyQt5.QtWidgets import *
else :
   from PyQt4.QtCore import *
   from PyQt4.QtGui import *

# from lexer     import Lexer, fileNameToIndex
from grammar   import Grammar # , Rule, Expression, Alternative, Ebnf, Nonterminal, Terminal
from toparser  import ToParser
from toproduct import ToProduct

from util import findColor
from area import Area
from support import Support, Sections

# --------------------------------------------------------------------------

class ExamplePlugin (QObject) :

   def __init__ (self, main_window) :
       super (ExamplePlugin, self).__init__ (main_window)
       self.win = main_window

       self.exampleMenu = self.win.menuBar().addMenu ("Exa&mple")

       act = QAction ("cecko.g", self.win)
       act.triggered.connect (self.test_cecko)
       self.exampleMenu.addAction (act)

       act = QAction ("cecko2.g", self.win)
       act.triggered.connect (self.test_cecko2)
       self.exampleMenu.addAction (act)

       act = QAction ("cecko2.g lout", self.win)
       act.triggered.connect (self.test_cecko2_lout)
       self.exampleMenu.addAction (act)

       act = QAction ("cecko3.g", self.win)
       act.setShortcut ("F10")
       act.triggered.connect (self.test_cecko3)
       self.exampleMenu.addAction (act)

       act = QAction ("cecko3.g diagram", self.win)
       act.setShortcut ("Shift+F10")
       act.triggered.connect (self.test_cecko3_diagram)
       self.exampleMenu.addAction (act)

       # act = QAction ("cecko4.g", self.win)
       # act.triggered.connect (self.test_cecko4)
       # self.exampleMenu.addAction (act)

   def test_cecko (self) :
       self.grammarFileName = "examples/cecko.g"
       self.sourceFileName = "examples/cecko.cc"
       self.test ()

   def test_cecko2 (self) :
       self.grammarFileName = "examples/cecko2.g"
       self.sourceFileName = "examples/cecko2.cc"
       result = self.test (run = True)

   def test_cecko2_lout (self) :
       self.grammarFileName = "examples/cecko2.g"
       self.sourceFileName = "examples/cecko2.cc"
       result = self.test (run = True)
       self.test_to_lout ("examples/cecko2_to_lout.py", result)

   def test_cecko3 (self) :
       self.grammarFileName = "examples/cecko3.g"
       self.sourceFileName = "examples/cecko3.cc"
       result = self.test (run = True, compilerFileName = "examples/cecko3_compiler.py")
       # self.test_to_python ("examples/cecko3_to_python.py", result)
       self.win.loadFile (self.sourceFileName)        # show source

   def test_cecko3_diagram (self) :
       self.grammarFileName = "examples/cecko3.g"
       self.sourceFileName = "examples/cecko3.cc"
       result = self.test (run = True, compilerFileName = "examples/cecko3_compiler.py")
       self.diagram (result)

   def test_cecko4 (self) :
       self.grammarFileName = "examples/cecko4.g"
       self.sourceFileName = "examples/cecko4.cc"
       result = self.test (run = True, compilerFileName = "examples/cecko4_compiler.py")

   # Test

   def test (self, run = False, compilerFileName = "") :

       shortName, ext = os.path.splitext (os.path.basename (self.grammarFileName))

       parserFileName = self.win.outputFileName (shortName + "_parser.py")
       productFileName = self.win.outputFileName (shortName + "_product.py")

       sourceEdit = self.win.inputFile (self.sourceFileName)
       self.win.initProject (sourceEdit)

       rebuild = True

       if ( rebuild or
            self.win.rebuildFile (self.grammarFileName, parserFileName) or
            self.win.rebuildFile (self.grammarFileName, productFileName) ) :

            # parser

            print ("generating parser")
            self.win.showStatus ("generating parser")

            self.win.inputFile (self.grammarFileName) # save if modified

            grammar = Grammar ()
            grammar.openFile (self.grammarFileName)

            parser_generator = ToParser ()
            parser_generator.open (parserFileName)
            parser_generator.parserFromGrammar (grammar)
            parser_generator.close ()

            grammarEdit = self.win.inputFile (self.grammarFileName)
            self.win.joinProject (grammarEdit)
            self.win.displayGrammarData (grammarEdit, grammar)

            parserEdit = self.win.reloadFile (parserFileName)
            self.win.joinProject (parserEdit)
            self.win.displayPythonCode (parserEdit)

            print ("parser O.K.")

            # product

            if run :
               parserModuleName, ext = os.path.splitext (os.path.basename (parserFileName))

               product_generator = ToProduct ()
               product_generator.open (productFileName)
               product_generator.productFromGrammar (grammar, parserModuleName)
               product_generator.close ()

               productEdit = self.win.reloadFile (productFileName)
               self.win.joinProject (productEdit)
               self.win.displayPythonCode (productEdit)

               print ("product O.K.")

            self.win.showStatus ("")

       result = None

       # run parser

       if run :
          parser_module = self.win.loadModule (parserFileName) # load parser module (required for compile_module)
          self.parser_module = parser_module # store for diagram drawing

          if compilerFileName == "" :
             parser_object = parser_module.Parser ()

             parser_object.openFile (self.sourceFileName)
             result = parser_object.parse_program ()
             parser_object.close ()

             self.win.displayCompilerData (sourceEdit, result)

          else :
             compiler_module = self.win.loadModule (compilerFileName)
             compiler_object = compiler_module.Compiler ()
             Support (compiler_object, sourceEdit)

             compiler_object.openFile (self.sourceFileName)
             result = compiler_object.parse_program ()
             compiler_object.close ()

             self.win.displayCompilerData (sourceEdit, result)
             self.win.showClasses (compiler_object.global_scope)
             self.win.addIdentifiers (sourceEdit, compiler_object.global_scope)

          print ("run parser O.K.")

       # run product

       if run :
          outputFileName = self.win.outputFileName (self.sourceFileName, "_output.cpp")

          product_module = self.win.loadModule (productFileName)
          product_object = product_module.Product ()

          product_object.open (outputFileName)
          product_object.send_program (result)
          product_object.close ()

          productEdit = self.win.reloadFile (outputFileName)
          self.win.joinProject (productEdit)
          self.win.displayFile (outputFileName)

          print ("run product O.K.")

       return result

   # Lout

   def test_to_lout (self, loutModuleFileName, result) :
       outputFileName = self.win.outputFileName (self.sourceFileName, "_input.lout")

       lout_module = self.win.loadModule (loutModuleFileName)
       lout_object = lout_module.ExampleToLout ()

       lout_object.open (outputFileName)
       lout_object.send_program (result)
       lout_object.close ()

       self.win.reloadFile (outputFileName)

       # create PDF or PostScript file

       subdir = os.path.dirname (outputFileName)
       save_dir = os.getcwd ()
       os.chdir (subdir)

       fileName = os.path.basename (outputFileName) # local file name

       psFileName, ext = os.path.splitext (fileName)
       psFileName = psFileName + ".ps"

       # yum install lout
       os.system ("lout" + " " +  fileName + " -o " + psFileName)
       # lout also creates .li file in current directory

       self.win.showPdf (psFileName)

       os.chdir (save_dir)
       print ("O.K.")

   # Python

   def test_to_python (self, generatorFileName, compiler_data) :
       outputFileName = self.win.outputFileName (self.sourceFileName, "_output.py")

       generator_module = self.win.loadModule (generatorFileName)
       generator_object = generator_module.C2Py ()

       outputEdit = self.win.newFile (outputFileName)
       self.win.joinProject (outputEdit)
       branch = self.win.displayFile (outputFileName)
       sections = Sections (branch, outputEdit)
       generator_object.setSections (sections)

       generator_object.open (outputFileName)
       generator_object.send_program (compiler_data)
       generator_object.close ()

       self.win.reloadFile (outputFileName)

       print ("O.K.")

   # Diagram

   def createDiagram (self, code, x0, y0) :
       area = Area ()
       code.area = area # store graphics object

       if isinstance (code, self.parser_module.TClass) :
          color = "cornflowerblue"
       elif isinstance (code, self.parser_module.TSimpleDecl) :
          if hasattr (code, "body") and code.body != None:
             color = "lime"
          else :
             color = "orange"
       elif isinstance (code, self.parser_module.TParamDecl) :
             color = "yellow"
       else :
          color = "wheat"

       color = findColor (color)
       area.setBrush (QBrush (color))

       area.setPen (QColor ("red"))

       xspace = 20
       yspace = 20
       x = x0 + xspace
       y = y0 + yspace
       w = 60 + 2 * xspace
       h = 2 * yspace
       if hasattr (code, "items") :
          for item in code.items :
              self.createDiagram (item, x, y)

              rect = item.area.boundingRect () # get sub-item size
              width = rect.width ()
              height = rect.height ()

              area.addArea (item.area) # place sub-item
              item.area.setRect (x, y, width, height) # locate sub-item

              if w < width + 2*xspace :
                 w = width + 2*xspace
              y = y + height + yspace
              h = h + height + yspace

       area.setRect (x0, y0, w, h) # update size

       if hasattr (code, "name") :
          area.setToolTip (code.name)

          f = QGraphicsSimpleTextItem ()
          f.setText (code.name)
          f.setPen (findColor ("red"))
          f.setBrush (findColor ("red"))
          f.setPos (area.rect().left()+10, area.rect().top()+5)
          f.setParentItem (area)

   def diagram (self, code) :
       self.createDiagram (code, 100, 100)
       designer = self.win.design # get graphical window
       designer.addArea (code.area) # show graphical object
       self.win.setFirstTab () # show graphical page

# --------------------------------------------------------------------------

# kate: indent-width 1; show-tabs true; replace-tabs true; remove-trailing-spaces all
