
/* statements */

while_stat < TWhileStat: TStat > :
   "while"
   "(" cond:expr ")"
   code:inner_stat ;

if_stat < TIfStat: TStat >  :
   "if"
   "(" cond:expr ")"
   then_code:inner_stat
   (
      <new_line>
      "else"
      else_code:inner_stat
   )?   ;

compound_stat < TCompoundStat: TSTat > :
   "{"
   <indent>
   code:stat_list
   <unindent>
   <new_line>
    "}" ;

simple_stat < TSimpleStat: TStat >:
   subexpr:expr
   ";" ;

empty_stat < TEmptyStat: TStat > :
   ";" ;

stat < select TStat > :
   while_stat | if_stat | compound_stat | simple_stat | empty_stat ;

inner_stat < return TStat > :
   <new_line>
   <indent>
   stat
   <unindent > ;

stat_list < TStatList > :
    (
       <new_line>
       <add> stat
    )* ;


/* -------------------------------------------------------------------- */

/* expressions */

< expression TExpr >
< struct TExpr { enum kind; } >
< struct TBinaryExpr : TExpr {  } >

variable_expr <TVariableExpr:TExpr, kind=varExp> :
   name:identifier
   <execute on_variable_expr> ;

int_value_expr <TIntValue:TExpr, kind=intValueExp> :
  value:number;

real_value_expr <TRealValue:TExpr, kind=realValueExp> :
  value:real_number;

char_value_expr <TCharValue:TExpr, kind=charValueExp> :
  value:character_literal;

string_value_expr <TStringValue:TExpr, kind=stringValueExp> :
  value:string_literal;

subexpr_expr <TSubexprExpr:TExpr, kind=subexprExp> :
   "(" subexpr:expr ")" ;

sequence_expr <TSequanceExpr:TExpr, kind=subexprExp> :
   "[" param:expr_list "]" ;

this_expr <TThisExpr:TExpr, kind=thisExp>:
  "this";

simple_expr <select TExpr> :
   variable_expr | int_value_expr | real_value_expr | char_value_expr | string_value_expr | subexpr_expr | sequence_expr | this_expr;

/* postfix expression */

postfix_expr <choose TExpr>:
  simple_expr
  (
    index_expr |
    // call_expr |
    compound_expr |
    field_expr |
    post_inc_expr |
    post_dec_expr
  )* ;

index_expr <TIndexExpr:TExpr, kind=indexExp>:
  <store left:TExpr>
  '[' param:expr_list ']' ;

call_expr <TCallExpr:TExpr, kind=callExp>:
  <store left:TExpr>
  '(' param:expr_list ')' ;

compound_expr <TCompoundExpr:TExpr, kind=compoundExp>:
  <store left:TExpr>
  '{' body:stat_list '}' ;

field_expr <TFieldExpr:TExpr, kind=fieldExp>:
  <store left:TExpr>
  <no_space> '.' <no_space>
  name:identifier
  <execute on_field_expr> ;

post_inc_expr <TPostIncExpr:TExpr, kind=postIncExp>:
  <store left:TExpr>
  "++" ;

post_dec_expr <TPostDecExpr:TExpr, kind=postDecExp>:
  <store left:TExpr>
  "--" ;

/* unary expression */

unary_expr <select TExpr>:
    inc_expr
  | dec_expr
  | deref_expr
  | addr_expr
  | plus_expr
  | minus_expr
  | bit_not_expr
  | log_not_expr
  | allocation_expr
  | deallocation_expr
  | postfix_expr ; 

inc_expr <TUnaryExpr:TExpr, kind=incExp>:
  "++" param:unary_expr;

dec_expr <TUnaryExpr:TExpr, kind=decExp>:
  "--" param:unary_expr;

deref_expr <TUnaryExpr:TExpr, kind=derefExp>:
  '*' param:unary_expr;

addr_expr <TUnaryExpr:TExpr, kind=addrExp>:
  '&' param:unary_expr;

plus_expr <TUnaryExpr:TExpr, kind=plusExp>:
  '+' param:unary_expr;

minus_expr <TUnaryExpr:TExpr, kind=minusExp>:
  '-' param:unary_expr;

bit_not_expr <TUnaryExpr:TExpr, kind=bitNotExp>:
  '~' param:unary_expr;

log_not_expr <TUnaryExpr:TExpr, kind=logNotExp>:
  '!' param:unary_expr;

allocation_expr <TNewExpr:TExpr, kind=newExp>:
  "new"
  type:identifier
  (
    '(' init:expr_list ')'
  )? ;

deallocation_expr <TDeleteExpr:TExpr, kind=deleteExp>:
  "delete"
  param:unary_expr ;


/* binary expressions */

multiplicative_expr <choose TExpr>:
  postfix_expr
  (
    <new TMulExpr:TBinaryExpr>
    <store left:TExpr>
    ( '*' <set kind=mulExp> |
      '/' <set kind=divExp> |
      '%' <set kind=modExp> )
    right:postfix_expr
  )* ;

additive_expr <choose TExpr>:
  multiplicative_expr
  (
    <new TAddExpr:TBinaryExpr>
    <store left:TExpr>
    ( '+' <set kind=addExp> |
      '-' <set kind=subExp> )
    right:multiplicative_expr
  )* ;

shift_expr <choose TExpr>:
  additive_expr
  (
    <new TShiftExpr:TBinaryExpr>
    <store left:TExpr>
    ( "<<" <set kind=shlExp> |
      ">>" <set kind=shrExp> )
    right:additive_expr
  )* ;

relational_expr <choose TExpr>:
  shift_expr
  (
    <new TRelExpr:TBinaryExpr>
    <store left:TExpr>
    ( '<'  <set kind=ltExp> |
      '>'  <set kind=gtExp> |
      "<=" <set kind=leExp> |
      ">=" <set kind=geExp> )
    right:shift_expr
  )* ;

equality_expr <choose TExpr>:
  relational_expr
  (
    <new TEqExpr:TBinaryExpr>
    <store left:TExpr>
    ( "==" <set kind=eqExp> |
      "!=" <set kind=neExp> )
    right:relational_expr
  )* ;

and_expr <choose TExpr>:
  equality_expr
  (
    <new TAndExpr:TBinaryExpr>
    <store left:TExpr>
    '&' <set kind=bitAndExp>
    right:equality_expr
  )* ;

exclusive_or_expr <choose TExpr>:
  and_expr
  (
    <new TXorExpr:TBinaryExpr>
    <store left:TExpr>
    '^' <set kind=bitXorExp>
    right:and_expr
  )* ;

inclusive_or_expr <choose TExpr>:
  exclusive_or_expr
  (
    <new TOrExpr:TBinaryExpr>
    <store left:TExpr>
    ( '|' <set kind=bitOrExp> )
    right:exclusive_or_expr
  )* ;

logical_and_expr <choose TExpr>:
  inclusive_or_expr
  (
    <new TAndAndExpr:TBinaryExpr>
    <store left:TExpr>
    "&&" <set kind=logAndExp>
    right:inclusive_or_expr
  )* ;

logical_or_expr <choose TExpr>:
  logical_and_expr
  (
    <new TOrOrExpr:TBinaryExpr>
    <store left:TExpr>
    "||" <set kind=logOrExp>
    right:logical_and_expr
  )* ;

assignment_expr <choose TExpr>:
  logical_or_expr
  (
    <new TAssignExpr:TBinaryExpr>
    <store left:TExpr>
    ( '=' <set kind=assignExp>     |
      '+=' <set kind=assignAddExp> |
      '-=' <set kind=assignSubExp> )
    right:assignment_expr
  )? ;

expr <return TExpr> :
   assignment_expr ;

expr_list <TExprList>:
  (
     <add> expr
     ( ',' <add> expr )*
  )? ;

/* -------------------------------------------------------------------- */

/* declarations */

class_decl <TClass:TDecl> :
   "class"
   name:identifier
   <execute on_class>
   <new_line>
   "{"
      <execute open_class>
      <indent>
      ( <new_line> <add> simple_decl ) *
      <unindent>
      <new_line>
      <execute close_class>
   "}"
   ( ";" <set semicolon=true> )? ;

enum_decl <TEnum:TDecl> :
   "enum"
   name:identifier
   <execute on_enum>
   "{"
      <execute open_enum>
      ( <new_line> <add> enum_item ) *
      <execute close_enum>
   "}"
   ( ";" <set semicolon=true> )? ;

enum_item <TEnumItem:TDecl> :
   name:identifier
   <execute on_enum_item> ;

simple_decl <TSimpleDecl:TDecl> :
   type:identifier
   name:identifier
   <execute on_simple_decl>
   (
      ";" <set variable=true>
   |
      "("
      ( <add> param_decl ( ',' <add> param_decl )* )?
      ")"
      <new_line>
      "{"
      <execute open_function>
      body:stat_list
      <execute close_function>
      "}"
      ( ";" <set semicolon=true> )?
   ) ;

param_decl <TParamDecl:TDecl> :
   type:identifier
   name:identifier
   <execute on_param_decl> ;

decl <select TDecl> :
   class_decl | enum_decl | simple_decl ;

declaration_list <TDeclList> :
   ( <new_line> <add> decl <empty_line> )* ;

program < return TDeclList >:
   declaration_list ;

